import re

from coalib.bears.LocalBear import LocalBear
from coalib.results.Result import Result, RESULT_SEVERITY
from coalib.results.Diff import Diff


def generate_spacing_diff(file, filename, line, line_number,
                          match_object, required_spacing):
    """
    Generate a diff for incorrectly spaced control or variable tags.

    :param match_object:     A Match object containing the groups ``open``,
                             ``close`` containing the opening and closing
                             delimiters of a Jinja2 tag and ``content``
                             containing everything in between.
    :param required_spacing: The number of spaces expected after the ``open``
                             delimiter and before the ``close`` delimiter
    """
    diff = Diff(file)

    content_before = line[:match_object.start('open')]
    content_after = line[match_object.end('close'):]

    spacing = ' ' * required_spacing
    replacement = (
        '{before}{open}{spacing}{content}{spacing}{close}{after}'.format(
            before=content_before,
            spacing=spacing,
            after=content_after,
            content=match_object.group('content').strip(),
            open=match_object.group('open'),
            close=match_object.group('close')))
    diff.change_line(
        line_number,
        line,
        replacement)
    return {filename: diff}


def generate_label_diff(file, filename, line, line_number,
                        match_object, expected_label):
    """
    Generates a diff for a missing or wrong control loop end label.
    Missing labels will be added, wrong ones replaced, content after the
    label will be left untouched.


    :param match_object:   A Match object containing the groups ``close``
                           containing the closing delimiters of a control
                           end tag and the optional group ``label``
                           containing the end tag label.
    :param expected_label: The expected label for that control block.

    """
    diff = Diff(file)

    content_before = line[:match_object.end('close')]

    suffix_start = match_object.end('close')
    # if a label is there, we cut it out, by starting the kept suffix after it
    if match_object.group('label') is not None:
        suffix_start = match_object.end('label')

    content_after = line[suffix_start:]

    replacement = '{before}{label}{after}'.format(
        before=content_before,
        after=content_after,
        label=expected_label
    )
    diff.change_line(
        line_number,
        line,
        replacement)
    return {filename: diff}


def has_required_spacing(string, required_spacing):
    """
    Checks if a string has the required amount of spaces
    on each side.

    :param string:           The string which is to be checked.
    :param required_spacing: The number of spaces expected on both
                             sides of the string.
    :return:                 True if the string has the required
                             number of spaces on both sides, False
                             otherwise.

    >>> has_required_spacing("  foo  ", 2)
    True
    >>> has_required_spacing(" foo", 1)
    False
    """
    leading_spaces = len(string) - len(string.lstrip(' '))
    trailing_spaces = len(string) - len(string.rstrip(' '))

    return (
        leading_spaces == required_spacing
        and trailing_spaces == required_spacing)


class Jinja2Bear(LocalBear):
    LANGUAGES = {'Jinja2'}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    ASCIINEMA_URL = 'https://asciinema.org/a/azi6u1gcxutoxn0l7xpu4pljp'
    CAN_DETECT = {'Syntax'}
    CAN_FIX = {'Formatting', 'Documentation'}

    VARIABLE_REGEX = re.compile(
        r'(?P<open>{{)(?P<content>.*?)(?P<close>}})')
    CONTROL_START_REGEX = re.compile(
        r'(?P<open>{%)(?P<content>\s*(for|if).*?)(?P<close>%})')
    CONTROL_END_REGEX = re.compile(
        r'(?P<open>{%)(?P<content>\s*end(for|if)\s*?)'
        r'(?P<close>%})(?P<label>{#.*?#})?')

    def handle_control_spacing_issue(self, file, filename, line, line_number,
                                     control_spacing, match_object):
        """
        Handles incorrectly spaced control tags (if and for),
        both start and end tags

        :param control_spacing: The number of spaces placed around control tags
                                as required by the settings.
        :param match_object:    A Match object containing the groups ``open``,
                                ``close`` containing the opening and closing
                                delimiters of a Jinja2 tag and ``content``
                                containing everything in between.
        """
        diff = generate_spacing_diff(
            file, filename, line, line_number, match_object, control_spacing)
        return Result.from_values(
            origin=self,
            message='Control blocks should be spaced with '
                    '`{0}` spaces on each side.'.format(
                        control_spacing),
            file=filename,
            line=line_number,
            column=match_object.start(0) + 1,
            end_line=line_number,
            end_column=match_object.end(0) + 1,
            diffs=diff)

    def check_for_variable_spacing_issues(self,
                                          file,
                                          filename,
                                          line,
                                          line_number,
                                          variable_spacing):
        """
        Checks any variable in the given line for spacing issues.
        Yields a Result for each issue found.

        :param file:
            The content of the file currently being inspected.
        :param filename:
            The name of the file currently being inspected.
        :param line:
            The content of the line currently being inspected.
        :param line_number:
            The current line number.
        :param variable_spacing:
            The number of spaces required on each side of a variable tag.
        """
        for m in self.VARIABLE_REGEX.finditer(line):
            match = m.group('content')
            if not has_required_spacing(match, variable_spacing):
                diff = generate_spacing_diff(
                    file, filename, line, line_number, m,
                    variable_spacing)
                yield Result.from_values(
                    origin=self,
                    message='Variable blocks should be spaced with '
                            '`{0}` spaces on each side.'.format(
                                variable_spacing),
                    file=filename,
                    line=line_number,
                    column=m.start(0) + 1,
                    end_line=line_number,
                    end_column=m.end(0) + 1,
                    diffs=diff)

    def check_control_start_tags(self,
                                 file,
                                 filename,
                                 line,
                                 line_number,
                                 control_spacing):
        """
        Checks any control start tag in the given line for spacing issues
        and puts the expected label on the ``control_stack``
        Yields a Result for each issue found.

        :param file:
            The content of the file currently being inspected.
        :param filename:
            The name of the file currently being inspected.
        :param line:
            The content of the line currently being inspected.
        :param line_number:
            The current line number.
        :param control_spacing:
            The number of spaces required on each side of a control tag.
        """
        for m in self.CONTROL_START_REGEX.finditer(line):
            # build the label which is expected at the end of this block
            end_label = '{{#{spacing}{content}{spacing}#}}'.format(
                content=m.group('content').strip(),
                spacing=' ' * control_spacing)
            self.control_stack.append((end_label, line_number))

            # check for spacing issues with this tag
            if not has_required_spacing(m.group('content'), control_spacing):
                yield self.handle_control_spacing_issue(
                    file, filename, line, line_number, control_spacing, m)

    def check_control_end_tags(self,
                               file,
                               filename,
                               line,
                               line_number,
                               control_spacing):
        """
        Checks any control end tag in the given line for spacing issues,
        missing/wrong labels or missing corresponding opening tag.
        Yields a Result for each issue found.

        :param file:
            The content of the file currently being inspected.
        :param filename:
            The name of the file currently being inspected.
        :param line:
            The content of the line currently being inspected.
        :param line_number:
            The current line number.
        :param control_spacing:
            The number of spaces required on each side of a control tag.
        """
        for m in self.CONTROL_END_REGEX.finditer(line):
            label = m.group('label')
            try:
                expected_label, start_in_line = self.control_stack.pop()
            except IndexError:
                # an end tag without an opening tag
                yield Result.from_values(
                    origin=self,
                    message='Control end tag has no corresponding start',
                    file=filename,
                    line=line_number,
                    column=m.start(0) + 1,
                    end_line=line_number,
                    end_column=m.end(0) + 1,
                    severity=RESULT_SEVERITY.MAJOR)
                # Continue to next line as we've detected a syntax error
                # and label checking can give false results
                continue

            # check for spacing issues with this tag
            if not has_required_spacing(m.group('content'), control_spacing):
                yield self.handle_control_spacing_issue(
                    file, filename, line, line_number, control_spacing, m)

            # yield results for incorrect or missing end labels
            if label is None and line_number != start_in_line:
                diff = generate_label_diff(
                    file, filename, line, line_number, m, expected_label)
                yield Result.from_values(
                    origin=self,
                    message='Unlabeled control end tag',
                    file=filename,
                    line=line_number,
                    column=m.start(0) + 1,
                    end_line=line_number,
                    end_column=m.end(0) + 1,
                    diffs=diff)
            elif label != expected_label and line_number != start_in_line:
                diff = generate_label_diff(
                    file, filename, line, line_number, m, expected_label)
                yield Result.from_values(
                    origin=self,
                    message='End tag label does not match expected label',
                    file=filename,
                    line=line_number,
                    column=m.start('label') + 1,
                    end_line=line_number,
                    end_column=m.end('label') + 1,
                    diffs=diff)

    def check_for_empty_control_stack(self, filename):
        """
        Checks if the control stack is empty and yields a Result for each
        item left on the stack.

        :param filename: The name of the file currently being inspected.
        """
        if len(self.control_stack) > 0:
            for (_, line_number) in self.control_stack:
                yield Result.from_values(
                    origin=self,
                    message='Control start tag has no corresponding end',
                    file=filename,
                    line=line_number,
                    severity=RESULT_SEVERITY.MAJOR)

    def run(self,
            filename,
            file,
            variable_spacing: int=1,
            control_spacing: int=1):
        """
        Check `Jinja2 templates <http://jinja.pocoo.org>`_ for syntax,
        formatting and documentation issues.
        The following aspects are being looked at:

        * Variable spacing:
            Variable tags should be padded with one space on each side, like
            this: ``{{ var_name }}``. This can be set to any number of spaces
            via the setting ``variable_spacing``.
            Malformatted variable tags are detected and fixes suggested.
        * Control spacing:
            Like variable spacing, but for control blocks, i.e. ``if`` and
            ``for`` constructs. Looks at both start and end block.
        * Control labels:
            It is good practice to label the end of an ``if`` or ``for``
            construct with a comment equal to the content of the start,
            like so::

                {% for x in y %}
                  do something
                {% endfor %}{# for x in y %}

            Mising or differing labels are detected and fixes suggested.
            Constructs with start and end on the same line are being ignored.
        * unbalanced blocks:
            Each opening tag for a ``for`` or ``if`` construct must be closed
            by a corresponding end tag. An unbalanced number of opening and
            closing tags is invalid syntax and will be reported with
            MAJOR severity by the bear.


        :param variable_spacing:
            The number of spaces a variable block should be spaced with.
            Default is 1.
        :param control_spacing:
            The number of spaces a control block should be spaced with.
            Default is 1.
        """

        # Whenever a control construct starts a tuple of the
        # expected end label and line number are put on this stack.
        # Whenever an end tag is encountered the last item added is popped.
        self.control_stack = []

        for line_number, line in enumerate(file, start=1):

            yield from self.check_for_variable_spacing_issues(
                file, filename, line, line_number, variable_spacing)

            yield from self.check_control_start_tags(
                file, filename, line, line_number, control_spacing)

            yield from self.check_control_end_tags(
                file, filename, line, line_number, control_spacing)

        # We've reached the end of the file.
        # Check if all control blocks have been closed
        yield from self.check_for_empty_control_stack(filename)



"""
coala-bears is a Python package containing all
the bears that are officially supported by coala.
"""


import os
import sys


VERSION_FILE = os.path.join(os.path.dirname(__file__), 'VERSION')
with open(VERSION_FILE, 'r') as ver:
    VERSION = ver.readline().strip()


__version__ = VERSION


def assert_supported_version():  # pragma: no cover
    if sys.version_info < (3, 4):
        print('coala supports only python 3.4 or later.')
        exit(4)



from coala_utils.string_processing.Core import unescaped_search_for
from coalib.bears.LocalBear import LocalBear
from coalib.bearlib import deprecate_settings
from coalib.bearlib.languages.LanguageDefinition import LanguageDefinition
from coalib.bearlib.spacing.SpacingHelper import SpacingHelper
from coalib.results.SourceRange import SourceRange
from coalib.results.Result import Result, RESULT_SEVERITY
from coalib.results.AbsolutePosition import AbsolutePosition
from coalib.results.Diff import Diff

from bears.general.AnnotationBear import AnnotationBear


class IndentationBear(LocalBear):

    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_FIX = {'Formatting'}
    BEAR_DEPS = {AnnotationBear}  # pragma: no cover

    @deprecate_settings(indent_size='tab_width')
    def run(self,
            filename,
            file,
            dependency_results: dict,
            language: str,
            use_spaces: bool=True,
            indent_size: int=SpacingHelper.DEFAULT_TAB_WIDTH,
            coalang_dir: str=None):
        """
        It is a generic indent bear, which looks for a start and end
        indent specifier, example: ``{ : }`` where "{" is the start indent
        specifier and "}" is the end indent specifier. If the end-specifier
        is not given, this bear looks for unindents within the code to
        correctly figure out indentation.

        It also figures out hanging indents and absolute indentation of
        function params or list elements.

        It does not however support  indents based on keywords yet.
        for example:

            if(something)
            does not get indented

        undergoes no change.

        WARNING: The IndentationBear is experimental right now, you can report
        any issues found to https://github.com/coala/coala-bears

        :param filename:
            Name of the file that needs to be checked.
        :param file:
            File that needs to be checked in the form of a list of strings.
        :param dependency_results:
            Results given by the AnnotationBear.
        :param language:
            Language to be used for indentation.
        :param use_spaces:
            Insert spaces instead of tabs for indentation.
        :param indent_size:
            Number of spaces per indentation level.
        :param coalang_dir:
            Full path of external directory containing the coalang
            file for language.
        """
        lang_settings_dict = LanguageDefinition(
            language, coalang_dir=coalang_dir)
        annotation_dict = dependency_results[AnnotationBear.name][0].contents
        # sometimes can't convert strings with ':' to dict correctly
        if ':' in dict(lang_settings_dict['indent_types']).keys():
            indent_types = dict(lang_settings_dict['indent_types'])
            indent_types[':'] = ''
        else:
            indent_types = dict(lang_settings_dict['indent_types'])

        encapsulators = (dict(lang_settings_dict['encapsulators']) if
                         'encapsulators' in lang_settings_dict else {})

        encaps_pos = []
        for encapsulator in encapsulators:
            encaps_pos += self.get_specified_block_range(
                file, filename,
                encapsulator, encapsulators[encapsulator],
                annotation_dict)
        encaps_pos = tuple(sorted(encaps_pos, key=lambda x: x.start.line))

        comments = dict(lang_settings_dict['comment_delimiter'])
        comments.update(
            dict(lang_settings_dict['multiline_comment_delimiters']))

        try:
            indent_levels = self.get_indent_levels(
                file, filename,
                indent_types, annotation_dict, encaps_pos, comments)
        # This happens only in case of unmatched indents or
        # ExpectedIndentError.
        except (UnmatchedIndentError, ExpectedIndentError) as e:
            yield Result(self, str(e), severity=RESULT_SEVERITY.MAJOR)
            return

        absolute_indent_levels = self.get_absolute_indent_of_range(
            file, filename,
            encaps_pos, annotation_dict)

        insert = ' '*indent_size if use_spaces else '\t'

        no_indent_file = self._get_no_indent_file(file)
        new_file = self._get_basic_indent_file(no_indent_file,
                                               indent_levels,
                                               insert)
        new_file = self._get_absolute_indent_file(new_file,
                                                  absolute_indent_levels,
                                                  indent_levels,
                                                  insert)
        if new_file != list(file):
            wholediff = Diff.from_string_arrays(file, new_file)
            for diff in wholediff.split_diff():
                yield Result(
                    self,
                    'The indentation could be changed to improve readability.',
                    severity=RESULT_SEVERITY.INFO,
                    affected_code=(diff.range(filename),),
                    diffs={filename: diff})

    def _get_no_indent_file(self, file):
        no_indent_file = [line.lstrip() if line.lstrip() else '\n'
                          for line_nr, line in enumerate(file)]
        return no_indent_file

    def _get_basic_indent_file(self, no_indent_file, indent_levels, insert):
        new_file = []
        for line_nr, line in enumerate(no_indent_file):
            new_file.append(insert*indent_levels[line_nr] + line
                            if line is not '\n' else '\n')

        return new_file

    def _get_absolute_indent_file(self,
                                  indented_file,
                                  absolute_indent_levels,
                                  indent_levels,
                                  insert):
        for _range, indent in absolute_indent_levels:
            prev_indent = get_indent_of_line(indented_file,
                                             _range.start.line - 1,
                                             length=False)
            prev_indent_level = indent_levels[_range.start.line - 1]
            for line in range(_range.start.line, _range.end.line):
                new_line = (prev_indent + ' '*indent +
                            insert*(indent_levels[line] - prev_indent_level) +
                            indented_file[line].lstrip())
                indented_file[line] = new_line if new_line.strip() != ''\
                    else '\n'
        return indented_file

    def get_absolute_indent_of_range(self,
                                     file,
                                     filename,
                                     encaps_pos,
                                     annotation_dict):
        """
        Gets the absolute indentation of all the encapsulators.

        :param file:            A tuple of strings.
        :param filename:        Name of file.
        :param encaps_pos:      A tuple ofSourceRanges of code regions
                                trapped in between a matching pair of
                                encapsulators.
        :param annotation_dict: A dictionary containing sourceranges of all the
                                strings and comments within a file.
        :return:                A tuple of tuples with first element as the
                                range of encapsulator and second element as the
                                indent of its elements.
        """
        indent_of_range = []
        for encaps in encaps_pos:
            indent = get_element_indent(file, encaps)
            indent_of_range.append((encaps, indent))

        return tuple(indent_of_range)

    def get_indent_levels(self,
                          file,
                          filename,
                          indent_types,
                          annotation_dict,
                          encapsulators,
                          comments):
        """
        Gets the level of indentation of each line.

        :param file:            File that needs to be checked in the form of
                                a list of strings.
        :param filename:        Name of the file that needs to be checked.
        :param indent_types:    A dictionary with keys as start of indent and
                                values as their corresponding closing indents.
        :param annotation_dict: A dictionary containing sourceranges of all the
                                strings and comments within a file.
        :param encapsulators:   A tuple of sourceranges of all encapsulators of
                                a language.
        :param comments:        A dict containing all the types of comment
                                specifiers in a language.
        :return:                A tuple containing the levels of indentation of
                                each line.
        """
        ranges = []
        for indent_specifier in indent_types:
            if indent_types[indent_specifier]:
                ranges += self.get_specified_block_range(
                    file, filename,
                    indent_specifier, indent_types[indent_specifier],
                    annotation_dict)
            else:
                ranges += self.get_unspecified_block_range(
                    file, filename,
                    indent_specifier, annotation_dict, encapsulators, comments)

        ranges = sorted(ranges, key=lambda x: x.start.line)
        indent_levels = []
        indent, next_indent = 0, 0
        for line in range(0, len(file)):
            indent = next_indent
            for _range in ranges:
                if _range.start.line == line + 1:
                    next_indent += 1

                first_ch = file[line].lstrip()[:1]
                if(_range.end.line == line + 1 and
                   first_ch in indent_types.values()):
                    indent -= 1
                    next_indent -= 1

                elif _range.end.line == line + 1:
                    next_indent -= 1
            indent_levels.append(indent)

        return tuple(indent_levels)

    def get_specified_block_range(self,
                                  file,
                                  filename,
                                  open_specifier,
                                  close_specifier,
                                  annotation_dict):
        """
        Gets a sourceranges of all the indentation blocks present inside the
        file.

        :param file:            File that needs to be checked in the form of
                                a list of strings.
        :param filename:        Name of the file that needs to be checked.
        :param open_specifier:  A character or string indicating that the
                                block has begun.
        :param close_specifier: A character or string indicating that the block
                                has ended.
        :param annotation_dict: A dictionary containing sourceranges of all the
                                strings and comments within a file.
        :return:                A tuple whith the first source range being
                                the range of the outermost indentation while
                                last being the range of the most
                                nested/innermost indentation.
                                Equal level indents appear in the order of
                                first encounter or left to right.
        """
        ranges = []

        open_pos = list(self.get_valid_sequences(
            file, open_specifier, annotation_dict))
        close_pos = list(self.get_valid_sequences(
            file, close_specifier, annotation_dict))

        number_of_encaps = len(open_pos)
        if number_of_encaps != len(close_pos):
            raise UnmatchedIndentError(open_specifier, close_specifier)

        if number_of_encaps == 0:
            return ()

        stack = []
        text = ''.join(file)
        open_counter = close_counter = position = 0
        op_limit = cl_limit = False
        for position in range(len(text)):
            if not op_limit:
                if open_pos[open_counter].position == position:
                    stack.append(open_pos[open_counter])
                    open_counter += 1
                    if open_counter == number_of_encaps:
                        op_limit = True

            if not cl_limit:
                if close_pos[close_counter].position == position:
                    try:
                        op = stack.pop()
                    except IndexError:
                        raise UnmatchedIndentError(open_specifier,
                                                   close_specifier)
                    ranges.append(SourceRange.from_values(
                        filename,
                        start_line=op.line,
                        start_column=op.column,
                        end_line=close_pos[close_counter].line,
                        end_column=close_pos[close_counter].column))
                    close_counter += 1
                    if close_counter == number_of_encaps:
                        cl_limit = True

        return tuple(ranges)

    def get_unspecified_block_range(self,
                                    file,
                                    filename,
                                    indent_specifier,
                                    annotation_dict,
                                    encapsulators,
                                    comments):
        """
        Gets the range of all blocks which do not have an un-indent specifer.

        :param file:             File that needs to be checked in the form of
                                 a list of strings.
        :param filename:         Name of the file that needs to be checked.
        :param indent_specifier: A character or string indicating that the
                                 indentation should begin.
        :param annotation_dict: A dictionary containing sourceranges of all the
                                strings and comments within a file.
        :param encapsulators:   A tuple of sourceranges of all encapsulators of
                                a language.
        :param comments:        A dict containing all the types of comments
                                specifiers in a language.
        :return:                A tuple of SourceRanges of blocks without
                                un-indent specifiers.
        """
        specifiers = list(self.get_valid_sequences(
            file,
            indent_specifier,
            annotation_dict,
            encapsulators,
            check_ending=True))
        _range = []
        for specifier in specifiers:
            current_line = specifier.line
            indent = get_indent_of_specifier(file, current_line, encapsulators)
            unindent_line = get_first_unindent(indent,
                                               file,
                                               current_line,
                                               annotation_dict,
                                               encapsulators,
                                               comments)

            if unindent_line == specifier.line:
                raise ExpectedIndentError(specifier.line)

            _range.append(SourceRange.from_values(
                filename,
                start_line=specifier.line, start_column=None,
                end_line=unindent_line, end_column=None))

        return tuple(_range)

    @staticmethod
    def get_valid_sequences(file,
                            sequence,
                            annotation_dict,
                            encapsulators=None,
                            check_ending=False):
        """
        A vaild sequence is a sequence that is outside of comments or strings.

        :param file:            File that needs to be checked in the form of
                                a list of strings.
        :param sequence:        Sequence whose validity is to be checked.
        :param annotation_dict: A dictionary containing sourceranges of all the
                                strings and comments within a file.
        :param encapsulators:   A tuple of SourceRanges of code regions
                                trapped in between a matching pair of
                                encapsulators.
        :param check_ending:    Check whether sequence falls at the end of the
                                line.
        :return:                A tuple of AbsolutePosition's of all occurances
                                of sequence outside of string's and comments.
        """
        file_string = ''.join(file)
        # tuple since order is important
        sequence_positions = tuple()

        for sequence_match in unescaped_search_for(sequence, file_string):
            valid = True
            sequence_position = AbsolutePosition(
                                    file, sequence_match.start())
            sequence_line_text = file[sequence_position.line - 1]

            # ignore if within string
            for string in annotation_dict['strings']:
                if(gt_eq(sequence_position, string.start) and
                   lt_eq(sequence_position, string.end)):
                    valid = False

            # ignore if within comments
            for comment in annotation_dict['comments']:
                if(gt_eq(sequence_position, comment.start) and
                   lt_eq(sequence_position, comment.end)):
                    valid = False

                if(comment.start.line == sequence_position.line and
                        comment.end.line == sequence_position.line and
                        check_ending):
                    sequence_line_text = sequence_line_text[
                        :comment.start.column - 1] + sequence_line_text[
                        comment.end.column-1:]

            if encapsulators:
                for encapsulator in encapsulators:
                    if(gt_eq(sequence_position, encapsulator.start) and
                       lt_eq(sequence_position, encapsulator.end)):
                        valid = False

            if not sequence_line_text.rstrip().endswith(':') and check_ending:
                valid = False

            if valid:
                sequence_positions += (sequence_position,)

        return sequence_positions


def get_indent_of_specifier(file, current_line, encapsulators):
    """
    Get indentation of the indent specifer itself.

    :param file:          A tuple of strings.
    :param current_line:  Line number of indent specifier (initial 1)
    :param encapsulators: A tuple with all the ranges of encapsulators
    :return:              Indentation of the specifier.
    """
    start = current_line
    _range = 0
    while (_range < len(encapsulators) and
           encapsulators[_range].end.line <= current_line):

        if current_line == encapsulators[_range].end.line:
            if encapsulators[_range].start.line < start:
                start = encapsulators[_range].start.line

        _range += 1

    return len(file[start - 1]) - len(file[start - 1].lstrip())


def get_first_unindent(indent,
                       file,
                       start_line,
                       annotation_dict,
                       encapsulators,
                       comments):
    """
    Get the first case of a valid unindentation.

    :param indent:          No. of spaces to check unindent against.
    :param file:            A tuple of strings.
    :param start_line:      The line from where to start searching for
                            unindent.
    :param annotation_dict: A dictionary containing sourceranges of all the
                            strings and comments within a file.
    :param encapsulators:   A tuple of SourceRanges of code regions trapped in
                            between a matching pair of encapsulators.
    :param comments:        A dict containing all the types of comments
                            specifiers in a language.
    :return:                The line where unindent is found (intial 0).
    """
    line_nr = start_line

    while line_nr < len(file):
        valid = True

        for comment in annotation_dict['comments']:
            if(comment.start.line < line_nr + 1 and
               comment.end.line >= line_nr + 1):
                valid = False

            first_char = file[line_nr].lstrip()[0] if file[line_nr].strip()\
                else ''
            if first_char in comments:
                valid = False

        for encapsulator in encapsulators:
            if(encapsulator.start.line < line_nr + 1 and
               encapsulator.end.line >= line_nr + 1):
                valid = False

        line_indent = len(file[line_nr]) - len(file[line_nr].lstrip())
        if line_indent <= indent and valid:
            return line_nr
        line_nr += 1

    return line_nr


# TODO remove these once https://github.com/coala/coala/issues/2377
# gets a fix
def lt_eq(absolute, source):
    if absolute.line == source.line:
        return absolute.column <= source.column

    return absolute.line < source.line


def gt_eq(absolute, source):
    if absolute.line == source.line:
        return absolute.column >= source.column

    return absolute.line > source.line


def get_element_indent(file, encaps):
    """
    Gets indent of elements inside encapsulator.

    :param file:   A tuple of strings.
    :param encaps: SourceRange of an encapsulator.
    :return:       The number of spaces params are indented relative to
                   the start line of encapsulator.
    """
    line_nr = encaps.start.line - 1
    start_indent = get_indent_of_line(file, line_nr)
    if len(file[line_nr].rstrip()) <= encaps.start.column:
        indent = get_indent_of_line(file, line_nr + 1)
    else:
        indent = encaps.start.column

    indent = indent - start_indent if indent > start_indent else 0
    return indent


def get_indent_of_line(file, line, length=True):
    indent = len(file[line]) - len(file[line].lstrip())
    if length:
        return indent
    else:
        return file[line][:indent]


class ExpectedIndentError(Exception):

    def __init__(self, line):
        Exception.__init__(self, 'Expected indent after line: ' + str(line))


class UnmatchedIndentError(Exception):

    def __init__(self, open_indent, close_indent):
        Exception.__init__(self, 'Unmatched ' + open_indent + ', ' +
                           close_indent + ' pair')



import re

from coalib.bears.LocalBear import LocalBear
from coalib.results.Result import Result


class PinRequirementsBear(LocalBear):

    LICENSE = 'AGPL'
    LANGUAGES = {
        'Python Requirements',
        'Python 2 Requirements',
        'Python 3 Requirements',
    }
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}

    def run(self, filename, file, require_patch: bool=False):
        """
        Checks if requirements are properly pinned. It will always raise an
        issue if the minor version is not given. If you do not wish that, do
        not use this bear.

        :param require_patch: Requires a patch version specified if True.
        """
        for i, line in enumerate(file, start=1):
            match = re.search(r'[=<>]+([0-9.]+)(?:$|\s|#)', line)
            if not match:
                continue

            if require_patch and len(match.group(0).split('.')) < 3:
                yield Result.from_values(
                    self, 'The patch version is missing.', filename, i)
            elif len(match.group(0).split('.')) < 2:
                yield Result.from_values(
                    self, 'The minor version is missing.', filename, i)



import platform
import shlex

from coalib.bearlib import deprecate_settings
from coalib.bearlib.abstractions.Linter import linter
from coalib.bearlib.spacing.SpacingHelper import SpacingHelper
from dependency_management.requirements.DistributionRequirement import (
    DistributionRequirement)


@linter(executable='indent' if platform.system() != 'Darwin' else 'gindent',
        use_stdin=True,
        output_format='corrected',
        result_message='Indentation can be improved.')
class GNUIndentBear:
    """
    This bear checks and corrects spacing and indentation via the well known
    Indent utility.

    C++ support is considered experimental.
    """

    LANGUAGES = {'C', 'C++'}
    REQUIREMENTS = {DistributionRequirement(apt_get='indent')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_FIX = {'Formatting'}

    @staticmethod
    @deprecate_settings(indent_size='tab_width')
    def create_arguments(filename, file, config_file,
                         max_line_length: int=79,
                         use_spaces: bool=True,
                         blank_lines_after_declarations: bool=False,
                         blank_lines_after_procedures: bool=False,
                         blank_lines_after_commas: bool=False,
                         braces_on_if_line: bool=False,
                         braces_on_func_def_line: bool=False,
                         cuddle_else: bool=False,
                         while_and_brace_on_same_line: bool=False,
                         case_indentation: int=0,
                         space_before_semicolon_after_empty_loop: bool=True,
                         delete_optional_blank_lines: bool=True,
                         declaration_indent: int=0,
                         brace_indent: int=2,
                         gnu_style: bool=False,
                         k_and_r_style: bool=False,
                         linux_style: bool=False,
                         indent_size: int=SpacingHelper.DEFAULT_TAB_WIDTH,
                         indent_cli_options: str=''):
        """
        :param max_line_length:
            Maximum number of characters for a line.
        :param use_spaces:
            True if spaces are to be used, else tabs.
        :param blank_lines_after_declarations:
            Forces blank lines after the declarations.

            Example:
                If ``blank_lines_after_declarations = True`` then
                ```
                int a;
                return ...;
                ```
                changes to
                ```
                int a;

                return ...;
                ```

        :param blank_lines_after_procedures:
            Force blank lines after procedure bodies.
        :param blank_lines_after_commas:
            Forces newline after comma in declaration.

            Example:
                If ``blank_lines_after_commas = True`` then
                ```
                int a, b;
                ```
                changes to
                ```
                int a,
                  b;
                ```
        :param braces_on_if_line:
            Puts the brace ``{`` on same line with if.

            Example:
                If ``braces_on_if_line = True``  then
                ```
                if (x > 0)
                {
                ```
                changes to
                ```
                if (x > 0) {
                ```
        :param braces_on_func_def_line:
            Puts the brace `{` on same line with the function declaration.
        :param cuddle_else:
            Cuddle else and preceding ``}``.

            Example:
                If ``cuddle_else = True`` then
                ```
                if (...) {
                    ....
                }
                else {
                ```
                changes to
                ```
                if (...) {
                    ....
                } else {
                ```

        :param while_and_brace_on_same_line:
            Cuddles while of ``do {} while``; and preceding ``}``.
        :param case_indentation:
            Specifies the number of spaces by which ``case`` in the ``switch``
            are indented.
        :param space_before_semicolon_after_empty_loop:
            Forces a blank before the semicolon ``;`` on one-line ``for`` and
            ``while`` statements.
        :param delete_optional_blank_lines:
             Deletes blank lines that are not needed. An example of needed
             blank line, is the blank line following a declaration when
             ``blank_line_after_declaration=True``.
        :param declaration_indent:
            Forces variables names to be aligned in column ``n`` with
            ``n = declaration_indent``  in declaration.

            Example:
                If ``declaration_indent = 8`` then,
                ```
                int a;
                float b;
                ```
                changes to
                ```
                int     a;
                float   b;
                ```
        :param brace_indent:
            Specifies the number of spaces by which braces are indented. Its
            default value is 2.
        :param gnu_style:
            Uses GNU coding style.
        :param k_and_r_style:
            Uses Kernighan & Ritchie coding style.
        :param linux_style:
            Uses Linux coding style.
        :param indent_size:
            Number of spaces per indentation level.
        :param indent_cli_options:
            Any command line options the indent binary understands. They
            will be simply passed through.
        """
        indent_options = ('--no-tabs' if use_spaces else '--use-tabs',
                          '--line-length', str(max_line_length),
                          '--indent-level', str(indent_size),
                          '--tab-size', str(indent_size), )
        indent_options += (('--cuddle-do-while',)
                           if while_and_brace_on_same_line
                           else ('--dont-cuddle-do-while',))
        indent_options += (('--swallow-optional-blank-lines',)
                           if delete_optional_blank_lines else ('-nsob',))
        indent_options += (('--blank-lines-after-declarations',)
                           if blank_lines_after_declarations else ('-nbad',))
        indent_options += (('--blank-lines-after-commas',)
                           if blank_lines_after_commas else ('-nbc',))
        indent_options += (('--blank-lines-after-procedures',)
                           if blank_lines_after_procedures else ('-nbap',))
        indent_options += (('-di' + str(declaration_indent),)
                           if declaration_indent != 0 else ())
        indent_options += (('--case-indentation'+str(case_indentation),)
                           if case_indentation != 0 else ())
        indent_options += (('--space-special-semicolon',)
                           if space_before_semicolon_after_empty_loop
                           else ('-nss',))
        indent_options += ('--brace-indent'+str(brace_indent),)
        indent_options += (('--braces-on-func-def-line',)
                           if braces_on_func_def_line else ('-blf',))
        indent_options += ((('-ce',) if cuddle_else else ('-nce',)) +
                           ('-br',)) if braces_on_if_line else ('-bl',)
        indent_style_option = ()
        indent_style_option += ('--gnu-style',) if gnu_style else ()
        indent_style_option += (('--k-and-r-style',)
                                if k_and_r_style and indent_style_option is ()
                                else ())
        indent_style_option += (('--linux-style',)
                                if linux_style and indent_style_option is ()
                                else ())
        # If a style is chosen the other configs aren't passed to `indent`
        return (indent_style_option if indent_style_option is not ()
                else indent_options) + tuple(shlex.split(indent_cli_options))



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.settings.Setting import typed_list


@linter(executable='pydocstyle',
        use_stdout=False,
        use_stderr=True,
        output_format='regex',
        output_regex=r'.*:(?P<line>\d+) .+:\n\s+(?P<message>.*)')
class PyDocStyleBear:
    """
    Checks python docstrings.
    """
    LANGUAGES = {'Python', 'Python 2', 'Python 3'}
    REQUIREMENTS = {PipRequirement('pydocstyle', '1.1')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Formatting', 'Documentation'}

    def create_arguments(self, filename, file, config_file,
                         pydocstyle_select: typed_list(str)=(),
                         pydocstyle_ignore: typed_list(str)=()):
        """
        :param pydocstyle_select:
            List of checked errors by specifying which errors to check for.
            Can't be used together with ``pydocstyle_ignore``.
        :param pydocstyle_ignore:
            List of checked errors by specifying which errors to ignore. Can't
            be used together with ``pydocstyle_select``.
        """
        args = (filename,)
        if pydocstyle_ignore and pydocstyle_select:
            self.err('The arguments pydocstyle_select and pydocstyle_ignore '
                     'are both given but mutually exclusive.')
            return
        elif pydocstyle_ignore:
            ignore = ','.join(part.strip() for part in pydocstyle_ignore)
            args += ('--ignore=' + ignore,)
        elif pydocstyle_select:
            select = ','.join(part.strip() for part in pydocstyle_select)
            args += ('--select=' + select,)

        return args



#!/usr/bin/env python3

import locale
import sys
from subprocess import call

import setuptools.command.build_py
from bears import Constants
from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand

try:
    locale.getlocale()
except (ValueError, UnicodeError):
    locale.setlocale(locale.LC_ALL, 'en_US.UTF-8')


class PyTestCommand(TestCommand):

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main([])
        sys.exit(errno)


class BuildDocsCommand(setuptools.command.build_py.build_py):
    apidoc_command = ('sphinx-apidoc', '-f', '-o', 'docs/API',
                      'bears')
    make_command = ('make', '-C', 'docs', 'html', 'SPHINXOPTS=-W')

    def run(self):
        err_no = call(self.apidoc_command)
        if not err_no:
            err_no = call(self.make_command)
        sys.exit(err_no)


with open('requirements.txt') as requirements:
    required = requirements.read().splitlines()
    required.remove('-r bear-requirements.txt')

with open('bear-requirements.txt') as requirements:
    bear_required = requirements.read().splitlines()

with open('test-requirements.txt') as requirements:
    test_required = requirements.read().splitlines()

with open('README.rst') as readme:
    long_description = readme.read()

extras_require = {
    'alldeps': bear_required,
}
required += [req for req in bear_required
             if not req.startswith('language-check')]


if __name__ == '__main__':
    setup(name='coala-bears',
          version=Constants.VERSION,
          description='Bears for coala (Code Analysis Application)',
          author='The coala developers',
          maintainer='Lasse Schuirmann, Fabian Neuschmidt, Mischa Kr\xfcger',
          maintainer_email=('lasse.schuirmann@gmail.com, '
                            'fabian@neuschmidt.de, '
                            'makman@alice.de'),
          url='http://coala.rtfd.org/',
          platforms='any',
          packages=find_packages(exclude=('build.*', 'tests', 'tests.*')),
          install_requires=required,
          extras_require=extras_require,
          tests_require=test_required,
          package_data={'bears': ['VERSION'],
                        'bears.java': ['checkstyle.jar', 'google_checks.xml'],
                        'bears.scala': ['scalastyle.jar',
                                        'scalastyle_config.xml']},
          license='AGPL-3.0',
          long_description=long_description,
          entry_points={'coalabears': ['coala_official_bears = bears']},
          # from http://pypi.python.org/pypi?%3Aaction=list_classifiers
          classifiers=[
              'Development Status :: 4 - Beta',

              'Environment :: Plugins',
              'Environment :: MacOS X',
              'Environment :: Win32 (MS Windows)',
              'Environment :: X11 Applications :: Gnome',

              'Intended Audience :: Science/Research',
              'Intended Audience :: Developers',

              'License :: OSI Approved :: GNU Affero General Public License '
              'v3 or later (AGPLv3+)',

              'Operating System :: OS Independent',

              'Programming Language :: Python :: Implementation :: CPython',
              'Programming Language :: Python :: 3.4',
              'Programming Language :: Python :: 3.5',
              'Programming Language :: Python :: 3 :: Only',

              'Topic :: Scientific/Engineering :: Information Analysis',
              'Topic :: Software Development :: Quality Assurance',
              'Topic :: Text Processing :: Linguistic'],
          cmdclass={'docs': BuildDocsCommand,
                    'test': PyTestCommand})



from coalib.bearlib import deprecate_settings
from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.NpmRequirement import NpmRequirement
from coala_utils.param_conversion import negate


@linter(executable='write-good',
        output_format='regex',
        output_regex=r'(?P<message>.*)\s*on\s*line\s*(?P<line>\d+)\s*at\s'
                      'column\s*(?P<column>\d+)'
        )
class WriteGoodLintBear:
    """
    Lints the text files using ``write-good`` for improving proses.

    See <https://github.com/btford/write-good> for more information.
    """
    LANGUAGES = {'Natural Language'}
    REQUIREMENTS = {NpmRequirement('write-good', '0.9.1')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    ASCIINEMA_URL = 'https://asciinema.org/a/80761'
    CAN_DETECT = {'Formatting', 'Grammar'}

    @staticmethod
    @deprecate_settings(allow_passive_voice=('check_passive_voice', negate),
                        allow_so_beginning=('check_so_beginning', negate),
                        allow_adverbs=('check_adverbs', negate),
                        allow_repeated_words=('check_repeated_words', negate),
                        allow_there_is=('check_there_is', negate),
                        allow_ambiguous_words=('check_ambiguos_words', negate),
                        allow_extra_words=('check_extra_words', negate),
                        allow_cliche_phrases=('check_cliche_exists', negate))
    def create_arguments(filename, file, config_file,
                         allow_passive_voice: bool=True,
                         allow_so_beginning: bool=True,
                         allow_adverbs: bool=True,
                         allow_repeated_words: bool=True,
                         allow_there_is: bool=True,
                         allow_ambiguous_words: bool=True,
                         allow_extra_words: bool=True,
                         allow_cliche_phrases: bool=True):
        """
        Using ``False`` will enable the check.

        :param allow_passive_voice:     Allows passive voice.
        :param allow_so_beginning:      Allows ``So`` at the beginning of
                                        the sentence.
        :param allow_adverbs:           Allows adverbs that can weaken the
                                        meaning, such as: ``really``,
                                        ``very``, ``extremely``, etc.
        :param allow_repeated_words:    Allows lexical illusions – cases
                                        where a word is repeated.
        :param allow_there_is:          Allows ``There is`` or ``There are``
                                        at the beginning of the sentence.
        :param allow_ambiguous_words:   Allows ``weasel words`` for example
                                        ``often``, ``probably``
        :param allow_extra_words:       Allows wordy phrases and unnecessary
                                        words.
        :param allow_cliche_phrases:    Allows common cliche phrases in the
                                        sentence.
        """
        arg_map = {
            'allow_passive_voice': '--passive',
            'allow_so_beginning': '--so',
            'allow_adverbs': '--adverb',
            'allow_repeated_words': '--illusion',
            'allow_there_is': '--thereIs',
            'allow_ambiguous_words': '--weasel',
            'allow_extra_words': '--tooWordy',
            'allow_cliche_phrases': '--cliches'
        }
        arg_values = locals()
        args = tuple(arg for key, arg in arg_map.items()
                     if not arg_values[key])
        return args + (filename,)



from os.path import abspath, dirname, join

from coalib.bearlib.abstractions.Linter import linter


scalastyle_config_file = join(dirname(abspath(__file__)),
                              'scalastyle_config.xml')


@linter(executable='java',
        output_format='regex',
        output_regex=r'(?P<severity>warning) file=.+ message=(?P<message>.+) '
                     r'line=(?P<line>\d+)(?: column=(?P<column>\d+))?')
class ScalaLintBear:
    """
    Check Scala code for codestyle, but also semantical problems,
    e.g. cyclomatic complexity.
    """

    LANGUAGES = {'Scala'}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    INCLUDE_LOCAL_FILES = {'scalastyle.jar', 'scalalint_config.xml'}
    CAN_DETECT = {'Formatting', 'Complexity'}

    def setup_dependencies(self):
        type(self).jar = self.download_cached_file(
            'https://oss.sonatype.org/content/repositories/releases/org/'
            'scalastyle/scalastyle_2.10/0.8.0/scalastyle_2.10-0.8.0-batch.jar',
            'scalastyle.jar')

    @staticmethod
    def create_arguments(filename, file, config_file,
                         scalalint_config: str=scalastyle_config_file):
        """
        :param scalalint_config: Path to a custom configuration file.
        """
        return ('-jar', ScalaLintBear.jar, filename, '--config',
                scalalint_config)



from coalib.bearlib.abstractions.Linter import linter
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY
from dependency_management.requirements.PipRequirement import PipRequirement


@linter(executable='dennis-cmd',
        output_format='regex',
        output_regex=r'(?P<message>(?P<severity>[EW])[0-9]{3}: .*)'
                     r'\n(?P<line>[0-9]+):.*\n(?P<end_line>[0-9]+):.*',
        severity_map={'W': RESULT_SEVERITY.NORMAL,
                      'E': RESULT_SEVERITY.MAJOR})
class DennisBear:
    """
    Lints your translated PO and POT files!

    Check multiple lint rules on all the strings in the PO file
    generating a list of errors and a list of warnings.

    See http://dennis.readthedocs.io/en/latest/linting.html for
    list of all error codes.

    http://dennis.readthedocs.io/
    """

    LANGUAGES = {'po', 'pot'}
    REQUIREMENTS = {PipRequirement('dennis', '0.8'),
                    # Workaround for https://github.com/willkg/dennis/issues/91
                    PipRequirement('click', '6.6')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Syntax'}

    @staticmethod
    def create_arguments(filename, file, config_file, allow_untranslated=True):
        """
        :param allow_untranslated: Set to false to display unchanged
                                   translation warning.
        """
        if allow_untranslated:
            return ('lint', filename, '--excluderules', 'W302')
        else:
            return ('lint', filename)



import subprocess
import sys

from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.NpmRequirement import NpmRequirement


@linter(executable='alex',
        output_format='regex',
        output_regex=r'(?P<line>\d+):(?P<column>\d+)-(?P<end_line>\d+):'
                     r'(?P<end_column>\d+)\s+(?P<severity>warning)\s+'
                     r'(?P<message>.+)')
class AlexBear:
    """
    Checks the markdown file with Alex - Catch insensitive, inconsiderate
    writing.

    Be aware that Alex and this bear only work on English text.
    For more information, consult <https://www.npmjs.com/package/alex>.
    """
    LANGUAGES = {'Natural Language'}
    REQUIREMENTS = {NpmRequirement('alex', '3')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'

    @classmethod
    def check_prerequisites(cls):
        parent_prereqs = super().check_prerequisites()
        if parent_prereqs is not True:  # pragma: no cover
            return parent_prereqs

        incorrect_pkg_msg = (
            'Please ensure that the package that has been installed is the '
            "one to 'Catch insensitive, inconsiderate writing'. This can be "
            'verified by running `alex --help` and seeing what it does.')
        try:
            output = subprocess.check_output(('alex', '--help'),
                                             stderr=subprocess.STDOUT)
        except (OSError, subprocess.CalledProcessError):
            return ('The `alex` package could not be verified. ' +
                    incorrect_pkg_msg)
        else:
            output = output.decode(sys.getfilesystemencoding())
            if 'Catch insensitive, inconsiderate writing' in output:
                return True
            else:
                return ("The `alex` package that's been installed seems to "
                        'be incorrect. ' + incorrect_pkg_msg)

    @staticmethod
    def create_arguments(filename, file, config_file):
        return filename,
